"""Fröling connect API Wrapper.

This library is an unofficial API wrapper for the Fröling Web Portal (https://connect-web.froeling.com/).
Some features are not yet implemented, like remote ignition for example.

Github and documentation: https://https://github.com/Layf21/froeling-connect.py
"""

from froeling.client import Froeling
from froeling.datamodels import (
    Address,
    Component,
    Facility,
    NotificationDetails,
    NotificationOverview,
    Parameter,
    UserData,
)
from froeling.session import Session

__all__ = [
    'Froeling',
    'Session',
    'Address',
    'Component',
    'Facility',
    'NotificationDetails',
    'NotificationOverview',
    'Parameter',
    'UserData',
]



"""Manages authentication, requests and error handling."""

import base64
import json
import logging
from collections.abc import Callable
from http import HTTPStatus
from typing import Any

from aiohttp import ClientSession
from aiohttp.typedefs import StrOrURL

from froeling import endpoints, exceptions

HTTP_STATUS_SUCCESS_MIN = 200
HTTP_STATUS_SUCCESS_MAX = 299


class Session:
    """Represents an authenticated session with the API.

    A `Session` manages authentication, tokens, and connection details
    required to communicate with the backend. It can be created with
    either a username/password combination or an existing token.
    Both user_id and token are set if the user is logged in.

    Attributes:
        user_id (int | None): ID of the authenticated user.
        token (str | None): Active authentication token for this session.

    """

    user_id: int | None = None
    token: str | None = None

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        token: str | None = None,
        *,
        auto_reauth: bool = False,
        token_callback: Callable[[str], Any] | None = None,
        lang: str = 'en',
        logger: logging.Logger | None = None,
        clientsession: ClientSession | None = None,
    ) -> None:
        """Initialize a new Session.

        You can authenticate by providing a `username` and `password`,
        or by supplying an existing `token`. When `auto_reauth` is enabled,
        the session will attempt to automatically refresh expired tokens.

        Args:
        ----
            username (str | None): Username for authentication.
            password (str | None): Password for authentication.
            token (str | None): Existing authentication token.
            auto_reauth (bool): Whether to automatically refresh tokens
                when they expire. Defaults to False.
            token_callback (Callable[[str], Any] | None): Optional function
                that is called whenever a new token is obtained.
            lang (str): Preferred language for API responses. Defaults to `'en'`.
            logger (logging.Logger | None): Logger instance for debugging and events.
            clientsession (ClientSession | None): Optional aiohttp
                client session to reuse instead of creating a new one.

        """
        if not (token or (username and password)):
            msg = 'Set either token or username and password.'
            raise ValueError(msg)
        if auto_reauth and not (username and password):
            msg = 'Set username and password to use auto_reauth.'
            raise ValueError(msg)

        self.clientsession = clientsession or ClientSession()
        self._headers = {'Accept-Language': lang}
        self.username = username
        self.password = password
        self.auto_reauth = auto_reauth
        self.token_callback = token_callback

        if token:
            self.set_token(token)

        self._logger = logger or logging.getLogger(__name__)
        self._reauth_previous = False  # Did the previous request result in renewing the token?

    async def close(self) -> None:
        """Close the session."""
        await self.clientsession.close()

    def set_token(self, token: str) -> None:
        """Set the token used in Authorization and updates/sets user-id.

        :param token The bearer token
        """
        self._headers['Authorization'] = token
        try:
            self.user_id = json.loads(base64.b64decode(token.split('.')[1] + '==').decode('utf-8'))['userId']
        except Exception as e:
            msg = 'Token is in an invalid format.'
            raise ValueError(msg) from e
        self.token = token

    async def login(self) -> dict:
        """Get a token using username and password.

        :return: Json sent by server (includes userdata)
        """
        data = {'osType': 'web', 'username': self.username, 'password': self.password}
        async with await self.clientsession.post(endpoints.LOGIN, json=data) as res:
            if not HTTP_STATUS_SUCCESS_MIN <= res.status <= HTTP_STATUS_SUCCESS_MAX:
                msg = f'Server returned {res.status}: "{await res.text()}"'
                raise exceptions.AuthenticationError(msg)
            token = res.headers['Authorization']
            if self.token_callback:
                self.token_callback(token)
            self.set_token(token)
            userdata = await res.json()
        self._logger.debug('Logged in with username and password.')
        return userdata

    async def request(self, method: str, url: StrOrURL, headers: dict | None = None, **kwargs: Any) -> Any:
        """Do a web request.

        :param method:
        :param url:
        :param headers: Additional headers used in the request
        :param kwargs:
        """
        self._logger.debug('Sent %s: %s', method.upper(), url)
        request_headers = self._headers
        if headers:
            request_headers |= headers

        try:
            async with await self.clientsession.request(method, url, headers=request_headers, **kwargs) as res:
                if HTTP_STATUS_SUCCESS_MIN <= res.status <= HTTP_STATUS_SUCCESS_MAX:
                    r = await res.text()
                    self._logger.debug('Got %s', r)
                    self._reauth_previous = False
                    return await res.json()

                if res.status == HTTPStatus.UNAUTHORIZED:
                    if self.auto_reauth:
                        if self._reauth_previous:
                            msg = 'Reauth did not work.'
                            raise exceptions.AuthenticationError(msg, await res.text())
                        self._logger.info('Error %s, renewing token...', await res.text())
                        await self.login()
                        self._logger.info('Reauthorized.')
                        self._reauth_previous = True
                        return await self.request(method, url, **kwargs)

                    self._logger.error('Request unauthorized')
                    msg = 'Request not authorized: '
                    raise exceptions.AuthenticationError(msg, await res.text())

                error_data = await res.text()
                msg = 'Unexpected return code'
                raise exceptions.NetworkError(
                    msg,
                    status=res.status,
                    url=res.url,
                    res=error_data,
                )

        except json.decoder.JSONDecodeError as e:
            raise exceptions.ParsingError(e.msg, e.doc, e.pos, url) from e



"""Test the Facility class."""

import pytest
from aioresponses import aioresponses
from froeling import Froeling, endpoints


@pytest.mark.asyncio
async def test_get_facility(load_json):
    facility_data = load_json('facility.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    with aioresponses() as m:
        m.get(endpoints.FACILITY.format(1234), status=200, payload=facility_data)

        async with Froeling(token=token) as api:
            f = await api.get_facilities()

            assert len(f) == 2
            f1, f2 = f

            assert f1.facility_id == 12345
            assert f1.equipment_number == 100321123
            assert f1.status == 'OK'
            assert f1.name == 'Facility name'
            assert f1.address.street == 'some street'
            assert f1.address.zip == '3210'
            assert f1.address.city == 'somewhere'
            assert f1.address.country == 'DE'
            assert f1.owner == 'Jimmy Smith'
            assert f1.role == 'OWNER'
            assert f1.favorite is False
            assert f1.allow_messages is True
            assert f1.subscribed_notifications is True
            assert (
                f1.picture_url
                == 'https://connect-api.froeling.com/aks/connect/v1.0/resources/service/user/1234/facility/12345/picture'
            )
            assert f1.protocol_3200_info['hoursSinceLastMaintenance'] == '1000'
            assert f1.protocol_3200_info['operationHours'] == '3000'
            assert f1.protocol_3200_info['active'] is False
            assert f1.protocol_3200_info['productType'] == 'T4e 230-250'
            assert f1.protocol_3200_info['status'] == 'OK'
            assert f1.facility_generation == 'GEN_3200'

            assert f2.facility_id == 54321
            assert f2.equipment_number == 100123321
            assert f2.status == 'OK'
            assert f2.name == 'Another facility name'
            assert f2.address.street == 'some street'
            assert f2.address.zip == '3210'
            assert f2.address.city == 'somewhere'
            assert f2.address.country == 'DE'
            assert f2.owner == 'Heisenberg'
            assert f2.role == 'OWNER'
            assert f2.favorite is False
            assert f2.allow_messages is True
            assert f2.subscribed_notifications is True
            assert (
                f2.picture_url
                == 'https://connect-api.froeling.com/aks/connect/v1.0/resources/service/user/1234/facility/12345/picture'
            )
            assert f2.protocol_3200_info['hoursSinceLastMaintenance'] == '2000'
            assert f2.protocol_3200_info['operationHours'] == '200'
            assert f2.protocol_3200_info['active'] is False
            assert f2.protocol_3200_info['productType'] == 'S4 Turbo 60'
            assert f2.protocol_3200_info['status'] == 'OK'
            assert f2.facility_generation == 'GEN_3200'


@pytest.mark.asyncio
async def test_get_facility_modified(load_json):
    facility_data = load_json('facility_modified.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    with aioresponses() as m:
        m.get(endpoints.FACILITY.format(1234), status=200, payload=facility_data)

        async with Froeling(token=token) as api:
            f = await api.get_facilities()
            assert len(f) == 3
            f1, f2, f3 = f

            assert f1.facility_id == 12345
            assert f1.equipment_number is None
            assert f1.status == 'OK'
            assert f1.name == 'Facility name'
            assert f1.address.street == 'some street'
            assert f1.address.zip == 'ABCD'
            assert f1.address.city == 'somewhere'
            assert f1.address.country == 'DE'
            assert f1.owner == 'Jimmy Smith'
            assert f1.role == 'OWNER'
            assert f1.favorite is False
            assert f1.allow_messages is True
            assert f1.subscribed_notifications is True
            assert f1.picture_url is None
            assert f1.protocol_3200_info is None
            assert f1.facility_generation == 'GEN_500'

            assert f2.facility_id == 17
            assert f2.equipment_number == 100123321
            assert f2.status == 'OK'
            assert f2.name == 'Another facility name'
            assert f2.address.street is None
            assert f2.address.zip is None
            assert f2.address.city is None
            assert f2.address.country is None
            assert f2.owner == 'Slippy Jim'
            assert f2.role == 'OWNER'
            assert f2.favorite is False
            assert f2.allow_messages is True
            assert f2.subscribed_notifications is True
            assert (
                f2.picture_url
                == 'https://connect-api.froeling.com/aks/connect/v1.0/resources/service/user/1234/facility/12345/picture'
            )
            assert f2.protocol_3200_info['hoursSinceLastMaintenance'] == '2000'
            assert f2.protocol_3200_info['operationHours'] == '200'
            assert f2.protocol_3200_info['active'] is False
            assert f2.protocol_3200_info['productType'] == 'AMG GT 63'
            assert f2.protocol_3200_info['status'] == 'BAD'
            assert f2.facility_generation == 'GEN_3200'

            assert f3.facility_id == 0
            assert f3.equipment_number == -567
            assert f3.status == 'NOTOK'
            assert f3.name == 'Another facility name'
            assert f3.address is None
            assert f3.owner == 'Slippy Jim'
            assert f3.role == 'OWNER'
            assert f3.favorite is False
            assert f3.allow_messages is True
            assert f3.subscribed_notifications is True
            assert (
                f3.picture_url
                == 'https://connect-api.froeling.com/aks/connect/v1.0/resources/service/user/1234/facility/12345/picture'
            )
            assert f3.protocol_3200_info is None
            assert f3.facility_generation == 'GEN_6400'


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'component_id,expected',
    [
        (
            '1_100',
            {
                'display_name': 'some display name',
                'display_category': 'Kessel',
                'standard_name': 'Kessel',
                'type': 'BOILER',
                'sub_type': 'WOODCHIP',
            },
        ),
        (
            '300_3100',
            {
                'display_name': 'Circuit 1',
                'display_category': 'Heizkreis',
                'standard_name': 'Heizkreis 01',
                'type': 'CIRCUIT',
                'sub_type': 'OUT_TEMP_CRTL',
            },
        ),
        (
            '300_3110',
            {
                'display_name': 'Circuit 2',
                'display_category': 'Heizkreis',
                'standard_name': 'Heizkreis 02',
                'type': 'CIRCUIT',
                'sub_type': 'OUT_TEMP_CRTL',
            },
        ),
        (
            '200_2100',
            {
                'display_name': 'Boiler 01',
                'display_category': 'Boiler',
                'standard_name': 'Boiler 01',
                'type': 'DHW',
                'sub_type': None,
            },
        ),
        (
            '400_4100',
            {
                'display_name': 'Puffer 01',
                'display_category': 'Puffer',
                'standard_name': 'Puffer 01',
                'type': 'BUFFER_TANK',
                'sub_type': 'NEW_GENERATION',
            },
        ),
    ],
)
async def test_facility_get_components(load_json, component_id, expected):
    facility_data = load_json('facility.json')
    component_list_data = load_json('component_list.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    with aioresponses() as m:
        m.get(endpoints.FACILITY.format(1234), status=200, payload=facility_data)
        m.get(
            endpoints.COMPONENT_LIST.format(1234, 12345),
            status=200,
            payload=component_list_data,
        )

        async with Froeling(token=token) as api:
            facilities = await api.get_facilities()
            components = await facilities[0].get_components()

            comp = next(c for c in components if c.component_id == component_id)

            for field, value in expected.items():
                assert getattr(comp, field) == value
                assert comp.time_windows_view is None
                assert comp.picture_url is None
                assert comp.parameters == {}


@pytest.mark.asyncio
async def test_facility_get_component(load_json):
    facility_data = load_json('facility.json')
    component_data = load_json('component.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    with aioresponses() as m:
        m.get(endpoints.FACILITY.format(1234), status=200, payload=facility_data)
        m.get(
            endpoints.COMPONENT.format(1234, 12345, '1_100'),
            status=200,
            payload=component_data,
        )
        m.get(
            endpoints.COMPONENT.format(1234, 12345, '1_100'),
            status=200,
            payload=component_data,
        )

        async with Froeling(token=token) as api:
            f = await api.get_facility(12345)
            c = f.get_component('1_100')
            c2 = api.get_component(12345, '1_100')
            assert c.component_id == c2.component_id



"""Dataclasses relating to Facilities."""

from dataclasses import dataclass

from froeling import endpoints
from froeling.datamodels.component import Component
from froeling.datamodels.generics import Address
from froeling.session import Session


@dataclass(frozen=True)
class Facility:
    """Represents data related to a facility."""

    session: Session
    facility_id: int
    equipment_number: int | None
    status: str | None
    name: str | None
    address: Address | None
    owner: str | None
    role: str | None
    favorite: bool | None
    allow_messages: bool | None
    subscribed_notifications: bool | None
    picture_url: str | None
    protocol_3200_info: dict[str, str] | None
    hours_since_last_maintenance: int | None
    operation_hours: int | None
    facility_generation: str | None

    @staticmethod
    def _from_dict(obj: dict, session: Session) -> 'Facility':
        facility_id = obj.get('facilityId')
        if not isinstance(facility_id, int):
            msg = f'facilityid was not an int.\nobj:{obj}'
            raise TypeError(msg)
        equipment_number = obj.get('equipmentNumber')
        status = obj.get('status')
        name = obj.get('name')
        address_data = obj.get('address')
        address = Address._from_dict(address_data) if isinstance(address_data, dict) else None  # noqa: SLF001
        owner = obj.get('owner')
        role = obj.get('role')
        favorite = obj.get('favorite')
        allow_messages = obj.get('allowMessages')
        subscribed_notifications = obj.get('subscribedNotifications')
        picture_url = obj.get('pictureUrl')

        protocol_3200_info = obj.get('protocol3200Info')

        hours_since_last_maintenance: int | None = None
        operation_hours: int | None = None
        if isinstance(protocol_3200_info, dict):
            hslm = protocol_3200_info.get('hoursSinceLastMaintenance')
            if isinstance(hslm, int | str):
                try:
                    hours_since_last_maintenance = int(hslm)
                except ValueError:
                    hours_since_last_maintenance = None

            op_hours = protocol_3200_info.get('operationHours')
            if isinstance(op_hours, int | str):
                try:
                    operation_hours = int(op_hours)
                except ValueError:
                    operation_hours = None

        facility_generation = obj.get('facilityGeneration')
        return Facility(
            session,
            facility_id,
            equipment_number,
            status,
            name,
            address,
            owner,
            role,
            favorite,
            allow_messages,
            subscribed_notifications,
            picture_url,
            protocol_3200_info,
            hours_since_last_maintenance,
            operation_hours,
            facility_generation,
        )

    @staticmethod
    def _from_list(obj: list, session: Session) -> list['Facility']:
        return [Facility._from_dict(i, session) for i in obj]

    async def get_components(self) -> list[Component | None]:
        """Fetch all components of this facility (not cached)."""
        res = await self.session.request(
            'get',
            endpoints.COMPONENT_LIST.format(self.session.user_id, self.facility_id),
        )
        return [Component._from_overview_data(self.facility_id, self.session, i) for i in res]  # noqa: SLF001

    def get_component(self, component_id: str) -> Component:
        """Get a component given it's id.

        Data will not be initialized, call the Component.update method to fetch them.
        """
        return Component(self.facility_id, component_id, self.session)



"""Test the FACILITY endpoint."""

import pytest
from http import HTTPStatus
from aioresponses import aioresponses
from froeling import Froeling, endpoints


@pytest.mark.asyncio
async def test_facility_get_components(load_json):
    facility_data = load_json('facility.json')
    component_list_data = load_json('component_list.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    with aioresponses() as m:
        m.get(endpoints.FACILITY.format(1234), status=200, payload=facility_data)
        m.get(
            endpoints.COMPONENT_LIST.format(1234, 12345),
            status=200,
            payload=component_list_data,
        )

        async with Froeling(token=token) as api:
            f = await api.get_facility(12345)
            c = await f.get_components()
            assert len(c) == 5
            c = c[0]

            assert c.component_id == '1_100'
            assert c.display_name == 'some display name'
            assert c.display_category == 'Kessel'
            assert c.standard_name == 'Kessel'
            assert c.type == 'BOILER'
            assert c.sub_type == 'WOODCHIP'


@pytest.mark.asyncio
async def test_component_update(load_json):
    component_data = load_json('component.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    with aioresponses() as m:
        m.get(
            endpoints.COMPONENT.format(1234, 12345, '1_100'),
            status=200,
            payload=component_data,
        )

        async with Froeling(token=token) as api:
            c = api.get_component(12345, '1_100')
            await c.update()
            for p in c.parameters.values():
                p.display_value
            # TODO: Add asserts


@pytest.mark.asyncio
async def test_component_set_value(load_json):
    component_data = load_json('component.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    with aioresponses() as m:
        m.get(
            endpoints.COMPONENT.format(1234, 12345, '1_100'),
            status=200,
            payload=component_data,
        )
        m.put(
            endpoints.SET_PARAMETER.format(1234, 12345, '3_0'),
            status=HTTPStatus.NOT_MODIFIED,
            payload='successmessage',
        )
        m.put(
            endpoints.SET_PARAMETER.format(1234, 12345, '3_0'),
            status=200,
            payload='successmessage',
        )

        async with Froeling(token=token) as api:
            c = api.get_component(12345, '1_100')
            await c.update()
            msg = await list(c.parameters.values())[0].set_value('testvalue')
            assert msg is None

            msg = await list(c.parameters.values())[0].set_value('testvalue')
            assert msg == 'successmessage'
